;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                              ;;
;; Copyright (C) KolibriOS team 2004-2012. All rights reserved. ;;
;; Distributed under terms of the GNU General Public License    ;;
;;                                                              ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

$Revision: 2511 $

; Old style system call converter
align 16
cross_order:
        ; load all registers in crossed order
        mov     eax, ebx
        mov     ebx, ecx
        mov     ecx, edx
        mov     edx, esi
        mov     esi, edi
        movzx   edi, byte[esp+28 + 4]
        sub     edi, 53
        call    dword [servetable+edi*4]
        ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                            ;;
;;                     SYSENTER ENTRY                         ;;
;;                                                            ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

align 32
sysenter_entry:
        ;  
        mov     esp, [ss:tss._esp0]
        sti
        push    ebp                     ; save app esp + 4
        mov     ebp, [ebp]              ; ebp - original ebp
        ;------------------
        pushad
        cld

        movzx   eax, al
        call    dword [servetable2 + eax * 4]

        popad
        ;------------------
        xchg    ecx, [ss:esp]           ;    - app ecx, ecx - app esp + 4
        sub     ecx, 4
        xchg    edx, [ecx]              ; edx - return point, & save original edx
        push    edx
        mov     edx, [ss:esp + 4]
        mov     [ecx + 4], edx          ; save original ecx
        pop     edx
        sysexit

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                            ;;
;;                   SYSTEM CALL ENTRY                        ;;
;;                                                            ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

align 16
i40:
        pushad
        cld
        movzx   eax, al
        call    dword [servetable2 + eax * 4]
        popad
        iretd

i80:
		pushad
		cld
		movzx   eax, al
		DEBUGF  1, "K : Unix call %d, ebx:%x ecx:%x edx:%x\n", eax, ebx, ecx, edx
        call    dword [nix_syscalls + eax * 4]
		popad
		iretd

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                            ;;
;;                     SYSCALL ENTRY                          ;;
;;                                                            ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
align 32
syscall_entry:
  ;     cli                 syscall clear IF
        xchg    esp, [ss:tss._esp0]
        push    ecx
        lea     ecx, [esp+4]
        xchg    ecx, [ss:tss._esp0]
        sti
        push    ecx
        mov     ecx, [ecx]
        ;------------------
        pushad
        cld

        movzx   eax, al
        call    dword [servetable2 + eax * 4]

        popad
        ;------------------
        mov     ecx, [ss:esp+4]
        pop     esp
        sysret

iglobal
  align 4
  nix_syscalls:

      dd unix_malloc     ;  0 - spec malloc
      dd unix_exit	           ;  1 - exit
      dd undefined_syscall     ;  2 - fork
      dd unix_read             ;  3 - read
      dd unix_write            ;  4 - write
      dd unix_open             ;  5 - open
      dd unix_close            ;  6 - close
      dd undefined_syscall     ;  7 - waitpid
      dd undefined_syscall     ;  8 - creat
      dd undefined_syscall     ;  9 - link
      dd undefined_syscall     ;  10 - unlink
      dd undefined_syscall     ;  11 - execve
      dd undefined_syscall     ;  12 - chdir
      dd unix_time             ;  13 - time
      dd undefined_syscall     ;  14 - mknod
      dd undefined_syscall     ;  15 - chmod
      dd undefined_syscall     ;  16 - lchown
      dd undefined_syscall     ;  not defined
      dd undefined_syscall     ;  18 - stat
      dd undefined_syscall     ;  19 - lseek
      dd undefined_syscall     ;  20 - getpid
      dd undefined_syscall     ;  21 - mount
      dd undefined_syscall     ;  22 - oldmount
      dd undefined_syscall     ;  23 - setuid
      dd undefined_syscall     ;  24 - getuid
      dd undefined_syscall     ;  25 - stime            
      dd undefined_syscall     ;  26 - ptrace
      dd undefined_syscall     ;  27 - alarm
      dd undefined_syscall     ;  28 - fstat
      dd undefined_syscall     ;  29 - pause
      dd undefined_syscall     ;  30 - utime
      dd undefined_syscall     ;  not defined
      dd undefined_syscall     ;  not defined
      dd undefined_syscall     ;  not defined
      dd undefined_syscall     ;  34 - nice
      dd undefined_syscall     ;  not defined
      dd undefined_syscall     ;  36 - sync
      dd undefined_syscall     ;  37 - kill
      dd undefined_syscall     ;  38 - rename 
      dd undefined_syscall     ;  39 - mkdir
      dd undefined_syscall     ;  40 - rmdir
      dd undefined_syscall     ;  41 - dup
      dd undefined_syscall     ;  42 - pipe
      dd undefined_syscall     ;  43 - times
      dd undefined_syscall     ;  not defined             
      dd undefined_syscall     ;  45 - brk
      dd undefined_syscall     ;  46 - setgid
      dd undefined_syscall     ;  47 - getgid
      dd undefined_syscall     ;  48 - signal
      dd undefined_syscall     ;  49 - geteuid
      dd undefined_syscall     ;  50 - getegid
      dd undefined_syscall     ;  51 - acct
      dd undefined_syscall     ;  52 - umount
      dd undefined_syscall     ;  not defined
      dd undefined_syscall     ;  54 - ioctl
      dd undefined_syscall     ;  55 - fnctl
      dd undefined_syscall     ;  not defined
      dd undefined_syscall     ;  57 - setpgid
      dd undefined_syscall     ;  not defined
      dd undefined_syscall     ;  59 - olduname
      dd undefined_syscall     ;  60 - umask
      dd undefined_syscall     ;  61 - chroot
      dd undefined_syscall     ;  62 - ustat            
      dd undefined_syscall     ;  63 - dup2
      dd undefined_syscall     ;  64 - getppid
      dd undefined_syscall     ;  65 - getpgrp
      dd undefined_syscall     ;  66 - setsid
      dd undefined_syscall     ;  67 - sigaction
      dd undefined_syscall     ;  68 - sgetmask
      dd undefined_syscall     ;  69 - ssetmask
      dd undefined_syscall     ;  70 - setreuid
      dd undefined_syscall     ;  71 - setregid
      dd undefined_syscall     ;  72 - sigsuspend
      dd undefined_syscall     ;  73 - sigpending
      dd undefined_syscall     ;  74 - sethostname
      dd undefined_syscall     ;  75 - setrlimit
      dd undefined_syscall     ;  76 - getrlimit
      dd undefined_syscall     ;  77 - getrusage
      dd undefined_syscall     ;  78 - gettimeofday
      dd undefined_syscall     ;  79 - settimeofday
      dd undefined_syscall     ;  80 - getgroups            
      dd undefined_syscall     ;  81 - setgroups
      dd undefined_syscall     ;  82 - oldselect
      dd undefined_syscall     ;  83 - symlink
      dd undefined_syscall     ;  84 - lstat
      dd undefined_syscall     ;  85 - readlink
      dd undefined_syscall     ;  86 - uselib 
      dd undefined_syscall     ;  87 - swapon
      dd undefined_syscall     ;  88 - reboot
      dd undefined_syscall     ;  89 - oldreaddir
      dd undefined_syscall     ;  90 - oldmmap
      dd undefined_syscall     ;  91 - munmap
      dd undefined_syscall     ;  92 - truncate
      dd undefined_syscall     ;  93 - ftruncate
      dd undefined_syscall     ;  94 - fchmod
      dd undefined_syscall     ;  95 - fchown
      dd undefined_syscall     ;  96 - getpriority
      dd undefined_syscall     ;  97 - setpriority
      dd undefined_syscall     ;  not defined            
      dd undefined_syscall     ;  99 - statfs
      dd undefined_syscall     ;  100 - fstatfs
      dd undefined_syscall     ;  101 - ioperm
      dd undefined_syscall     ;  102 - socketcall
      dd undefined_syscall     ;  103 - syslog
      dd undefined_syscall     ;  104 - settimer
      dd undefined_syscall     ;  105 - gettimer
      dd undefined_syscall     ;  106 - newstat
      dd undefined_syscall     ;  107 - newlstat
      dd undefined_syscall     ;  108 - newfstat
      dd undefined_syscall     ;  109 - uname
      dd undefined_syscall     ;  110 - iopl
      dd undefined_syscall     ;  111 - vhangup
      dd undefined_syscall     ;  112 - idle
      dd undefined_syscall     ;  113 - vm86old
      dd undefined_syscall     ;  114 - wait4
      dd undefined_syscall     ;  115 - swapoff
      dd undefined_syscall     ;  116 - sysinfo           
      dd undefined_syscall     ;  117 - ipc
      dd undefined_syscall     ;  118 - fsync
      dd undefined_syscall     ;  119 - sigreturn
      dd undefined_syscall     ;  120 - clone
      dd undefined_syscall     ;  121 - setdomainname
      dd undefined_syscall     ;  122 - newuname
      dd undefined_syscall     ;  123 - modifyldt
      dd undefined_syscall     ;  124 - adjtimex
      dd undefined_syscall     ;  125 - mprotect
      dd undefined_syscall     ;  126 - sigprocmask
      dd undefined_syscall     ;  127 - create_module
      dd undefined_syscall     ;  128 - init_module
      dd undefined_syscall     ;  129 - delete_module
      dd undefined_syscall     ;  130 - get_kernel_syms
      dd undefined_syscall     ;  131 - quotactl
      dd undefined_syscall     ;  132 - getpgid
      dd undefined_syscall     ;  133 - fchdir
      dd undefined_syscall     ;  134 - bdflush            
      dd undefined_syscall     ;  135 - sysfs
      dd undefined_syscall     ;  136 - personality
      dd undefined_syscall     ;  not defined
      dd undefined_syscall     ;  138 - setfsuid
      dd undefined_syscall     ;  139 - setfsgid
      dd undefined_syscall     ;  140 - llseek
      dd undefined_syscall     ;  141 - getdents
      dd undefined_syscall     ;  142 - select
      dd undefined_syscall     ;  143 - flock
      dd undefined_syscall     ;  144 - msync
      dd undefined_syscall     ;  145 - readv
      dd undefined_syscall     ;  146 - writev
      dd undefined_syscall     ;  147 - getsid
      dd undefined_syscall     ;  148 - fdatasync
      dd undefined_syscall     ;  149 - sysctl
      dd undefined_syscall     ;  150 - mlock
      dd undefined_syscall     ;  151 - munlock
      dd undefined_syscall     ;  152 - mlockall            
      dd undefined_syscall     ;  153 - munlockall
      dd undefined_syscall     ;  154 - sched_setparam
      dd undefined_syscall     ;  155 - sched_getparam
      dd undefined_syscall     ;  156 - sched_setsched
      dd undefined_syscall     ;  157 - sched_getsched
      dd undefined_syscall     ;  158 - sched_yield
      dd undefined_syscall     ;  159
      dd undefined_syscall     ;  160
      dd undefined_syscall     ;  161
      dd undefined_syscall     ;  162
      dd undefined_syscall     ;  163
      dd undefined_syscall     ;  164
      dd undefined_syscall     ;  165
      dd undefined_syscall     ;  166
      dd undefined_syscall     ;  167
      dd undefined_syscall     ;  168
      dd undefined_syscall     ;  169
      dd undefined_syscall     ;  170
      dd undefined_syscall     ;  171
      dd undefined_syscall     ;  172
      dd undefined_syscall     ;  173
      dd undefined_syscall     ;  174
      dd undefined_syscall     ;  175
      dd undefined_syscall     ;  176
      dd undefined_syscall     ;  177
      dd undefined_syscall     ;  178
      dd undefined_syscall     ;  179
      dd undefined_syscall     ;  180
      dd undefined_syscall     ;  181
      dd undefined_syscall     ;  182
      dd unix_getcwd           ;  183 - getcwd
      dd undefined_syscall     ;  184
      dd undefined_syscall     ;  185
      dd undefined_syscall     ;  186
      dd undefined_syscall     ;  187
      dd undefined_syscall     ;  188
      dd undefined_syscall     ;  189
      dd undefined_syscall     ;  190



  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;; SYSTEM FUNCTIONS TABLE ;;
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;

  align 4
  servetable:
      dd socket                  ; 53-Socket interface
      dd 0
      dd 0
      dd 0
      dd 0
      dd file_system             ; 58-Common file system interface
      dd 0
      dd 0
      dd 0
      dd 0                       ; 62-PCI functions
      dd sys_msg_board           ; 63-System message board

  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;; NEW SYSTEM FUNCTIONS TABLE ;;
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  align 4
  servetable2:

      dd syscall_draw_window     ; 0-DrawWindow
      dd syscall_setpixel        ; 1-SetPixel
      dd sys_getkey              ; 2-GetKey
      dd sys_clock               ; 3-GetTime
      dd syscall_writetext       ; 4-WriteText
      dd delay_hs                ; 5-DelayHs
      dd syscall_openramdiskfile ; 6-OpenRamdiskFile
      dd syscall_putimage        ; 7-PutImage
      dd syscall_button          ; 8-DefineButton
      dd sys_cpuusage            ; 9-GetProcessInfo
      dd sys_waitforevent        ; 10-WaitForEvent
      dd sys_getevent            ; 11-CheckForEvent
      dd sys_redrawstat          ; 12-BeginDraw and EndDraw
      dd syscall_drawrect        ; 13-DrawRect
      dd syscall_getscreensize   ; 14-GetScreenSize
      dd sys_background          ; 15-bgr
      dd sys_cachetodiskette     ; 16-FlushFloppyCache
      dd sys_getbutton           ; 17-GetButton
      dd sys_system              ; 18-System Services
      dd paleholder              ; 19-reserved
      dd sys_midi                ; 20-ResetMidi and OutputMidi
      dd sys_setup               ; 21-SetMidiBase,SetKeymap,SetShiftKeymap,.
      dd sys_settime             ; 22-setting date,time,clock and alarm-clock
      dd sys_wait_event_timeout  ; 23-TimeOutWaitForEvent
      dd syscall_cdaudio         ; 24-PlayCdTrack,StopCd and GetCdPlaylist
      dd syscall_putarea_backgr  ; 25-Put Area to background
      dd sys_getsetup            ; 26-GetMidiBase,GetKeymap,GetShiftKeymap,.
      dd undefined_syscall       ; 27-reserved
      dd undefined_syscall       ; 28-reserved
      dd sys_date                ; 29-GetDate
      dd sys_current_directory   ; 30-Get/SetCurrentDirectory
      dd undefined_syscall       ; 31-reserved
      dd undefined_syscall       ; 32-reserved
      dd undefined_syscall       ; 33-reserved
      dd syscall_getpixel_WinMap ; 34-GetPixel WinMap
      dd syscall_getpixel        ; 35-GetPixel
      dd syscall_getarea         ; 36-GetArea
      dd readmousepos            ; 37-GetMousePosition_ScreenRelative,.
      dd syscall_drawline        ; 38-DrawLine
      dd sys_getbackground       ; 39-GetBackgroundSize,ReadBgrData,.
      dd set_app_param           ; 40-WantEvents
      dd undefined_syscall       ; 41- deprecated GetIrqOwner
      dd undefined_syscall       ; 42- deprecated ReadIrqData
      dd sys_outport             ; 43-SendDeviceData
      dd undefined_syscall       ; 44- deprecated ProgramIrqs
      dd undefined_syscall       ; 45- deprecated ReserveIrq and FreeIrq
      dd syscall_reserveportarea ; 46-ReservePortArea and FreePortArea
      dd display_number          ; 47-WriteNum
      dd syscall_display_settings ; 48-SetRedrawType and SetButtonType
      dd sys_apm                 ; 49-Advanced Power Management (APM)
      dd syscall_set_window_shape ; 50-Window shape & scale
      dd syscall_threads         ; 51-Threads
      dd stack_driver_stat       ; 52-Stack driver status
      dd cross_order             ; 53-Socket interface
      dd undefined_syscall       ; 54-reserved
      dd sound_interface         ; 55-Sound interface
      dd undefined_syscall       ; 56-reserved
      dd sys_pcibios             ; 57-PCI BIOS32
      dd cross_order             ; 58-Common file system interface
      dd undefined_syscall       ; 59-reserved
      dd sys_IPC                 ; 60-Inter Process Communication
      dd sys_gs                  ; 61-Direct graphics access
      dd pci_api        ;cross_order             ; 62-PCI functions
      dd cross_order             ; 63-System message board
      dd sys_resize_app_memory   ; 64-Resize application memory usage
      dd sys_putimage_palette    ; 65-PutImagePalette
      dd sys_process_def         ; 66-Process definitions - keyboard
      dd syscall_move_window     ; 67-Window move or resize
      dd f68                     ; 68-Some internal services
      dd sys_debug_services      ; 69-Debug
      dd file_system_lfn         ; 70-Common file system interface, version 2
      dd syscall_window_settings ; 71-Window settings
      dd sys_sendwindowmsg       ; 72-Send window message
      dd blit_32                 ; 73-blitter;
      dd undefined_syscall       ; 74-reserved for new stack
      dd undefined_syscall       ; 75-reserved for new stack
      dd undefined_syscall       ; 76-reserved for new stack
        times 255 - ( ($-servetable2) /4 )  dd undefined_syscall
      dd sys_end                 ; -1-end application

endg
